/*
Copyright (c) 2006-2007, Group 1A TI ProHES 2006/2007
                         Fontys University of Professional Education
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @file bliepdsp.h
 * @brief BliepDSP Main API.
 *
 * BliepDSP backend for editing audio,
 * I/O and advance audio filtering.
 *
 * ToDo:
 *  - Many 24 and 32 bit functionality implementations.
 *  - Optimize filters to use the precaclulated sample amount.
 *  - Use circular smallbuf (IBM hack) wherever possible.
 *  - Child control and SMP support for POSIX like OSes. (making it less portable, Win32 lacks fork() and friends)
 *  - Progress indicator support.
 *  - Proper error handling.
 *  - Bit- and samplerate conversion.
 *  - More filters like FFT and Normalize.
 */

#ifndef BLIEPDSP_H
#define BLIEPDSP_H

/** @brief PI is needed. */
#define PI 3.141592653
/**
 * @brief id = RIFF.
 *
 * Values converted from big endian.
 */
#define PCM_RIFF 0x46464952
/**
 * @brief format = WAVE.
 *
 * Values converted from big endian.
 */
#define PCM_FORMAT 0x45564157
/** @brief Recursive filter parameter: HPF. */
#define HIGH_PASS 0xF0
/** @brief Recursive filter parameter: LPF. */
#define LOW_PASS 0xF1
/** @brief Recursive filter parameter: NPF. */
#define NARROW_PASS 0xF2
/** @brief Recursive filter parameter: NRF. */
#define NARROW_REJECT 0xF3

/**
 * @brief BliepDSP Audio Sample
 *
 * Every audio sample is stored withing an object like this.
 * Filters will be called over the object and another object emerges.
 * You can edit parts of an object by using the cut and paste functions.
 *
 * @warning All editing is done in memory so be sure to free() unused objects.
 */
class bliepdsp {
  public:
  char *data; /**< Pointer to actual audio data. */
  unsigned int num_samples; /**< Number of multichannel samples is precalculated, reduces math overhead. */

  /**
   * @brief PCM WAV header chunk
   *
   * The standard RIFF PCM WAV header defined by Microsoft.
   * Ironically Microsoft's own software doesn't reproduce this.
   */
  struct wavheader {
    unsigned int id; /**< Should be "RIFF". */
    unsigned int size; /**< Size of the first subchunk. */
    unsigned int format; /**< Should be "WAV". */
    unsigned int sub_id; /**< Should be "fmt ". */
    unsigned int sub_size; /**< Size of the second subchunk. */
    unsigned short int audio_format; /**< Should be 1 (PCM). */
    unsigned short int num_channels; /**< Number of channels */
    unsigned int sample_rate; /**< Sample rate in samples/sec. */
    unsigned int byte_rate; /**< Like we can't calculate the byterate for ourselves. Unused. */
    unsigned short int block_align; /**< Another unused block, same reason. Unused. */
    unsigned short int bits_per_sample; /**< Number of bits per audio sample. */
    unsigned int data_id; /**< Should be "data". */
    unsigned int data_size; /**< Size in bytes of the actual audio data. */
  } header;

  /**
   * @brief Calculate play time of current object.
   *
   * Returns seconds.
   */
  unsigned int calculate_time (void);

  /**
   * @brief Read WAV file in buffer.
   *
   * Object automatically mallocs memory.
   *
   * @param filename String filename pointing to the WAV file.
   * @warning Do not do this twice.
   */
  void read_wav (const char *filename);

  /**
   * @brief Write out WAV file.
   *
   * Writes current object to PCM WAV file.
   * @param filename String filename pointing to the WAV file to be written.
   */
  void write_wav (const char *filename);

  /**
   * @brief Free the buffer.
   *
   * Frees malloc'd memory when your program doesn't need it anymore.
   * You still have to tell it when to free the data though.
   */
  void freedata (void);

  /**
   * @brief Echo Filter.
   *
   * Echo filter with programmable delay and programmable wet value.
   * 8, 16 bit uncompressed PCM WAV.
   * 1-65535 channel audio.
   * @param out Object containing the actual result of the echo.
   * @param delay Delay in milliseconds for the echo.
   * @param wet Wet value for the echo. More wet means more echo.
   */
  void filter_echo (bliepdsp *out, const unsigned int delay, const unsigned int wet);

  /**
   * @brief Reverb Filter.
   *
   * Reverb filter with programmable delay, max impedance and programmable wet value.
   * 8, 16 bit uncompressed PCM WAV.
   * 1-65535 channel audio.
   * @param out Object containing the actual result of the reverb.
   * @param delay Delay in milliseconds for the reverberation part.
   * @param wet Wet value for the echo. More wet means more echo.
   * @param minimp Percentual impedance value on where to stop the filter.
   */
  void filter_reverb (bliepdsp *out, const unsigned int delay, const unsigned int wet, const unsigned int minimp);

  /**
   * @brief Backwards Filter.
   *
   * Backwards filter, puts sound buffer in reverse.
   * 8, 16, (24, 32)* bit PCM WAV.
   * 1-65535 channel audio.
   * * untested
   * @param out Object containing the actual result of the reverse.
   */
  void filter_backwards (bliepdsp *out);

  /**
   * @brief Stretch Filter.
   *
   * Stretch filter, stretches the sound buffer without making the bitrate and sample
   * rate change with programmable stretch.
   * 8, 16, (24, 32)* bit PCM WAV.
   * 1-65535 channel audio.
   * * untested.
   * @param out Object containing the actual stretched sample.
   * @param stretch Stretch in %: not 100, <100 shorten, >100 stretch.
  */
  void filter_stretch (bliepdsp *out, const unsigned int stretch);

  /**
   * @brief Recursive Multifunctional Filter.
   *
   * Recursive (single pole) all-in-one filter.
   * 8, 16 bit PCM WAV.
   * 1-65535 channel audio.
   * Currently supports:
   *  - HIGH_PASS (co is amount of low suppressed, 1-90%)
   *  - LOW_PASS (co is amount of high suppressed, 1-90%)
   *  - NARROW_PASS (co is frequency to pass, Hz, max half the sample rate)
   *  - NARROW_REJECT (co is frequency to reject, Hz, max half the sample rate)
   *
   * @param out Object contains the actual filtered sample.
   * @param filter One of the defined filter types.
   * @param co Variable to tune the filter behaviour.
   */
  void bliepdsp::filter_recursive (bliepdsp *out, const unsigned int filter, const unsigned int co);

  /**
   * @brief Copy a part from the stream.
   *
   * Parameters are multi-channel sample numbers.
   * 8, 16, (24, 32)* bit PCM WAV.
   * 1-65535 channel audio.
   * * untested.
   * @param out Object contains the copied part of the stream.
   * @param start Sample number to start copying from.
   * @param stop Sample number to stop copying at.
  */
  void bliepdsp::copy (bliepdsp *out, const unsigned int start, const unsigned int stop);

  /**
   * @brief Cut a part from the stream.
   *
   * The current object reduces in size.
   * Parameters are multi-channel sample numbers.
   * 8, 16, (24, 32)* bit PCM WAV.
   * 1-65535 channel audio.
   * * untested
   * @param out Object contains the cut out part of the stream.
   * @param start Sample number to start cutting from.
   * @param stop Sample number to stop cutting at.
  */
  void bliepdsp::cut (bliepdsp *out, const unsigned int start, const unsigned int stop);

  /**
   * @brief Paste sample into stream.
   *
   * Pastes a sample into the stream causing current stream to grow larger.
   * Parameter is a multi-channel sample number.
   * 8, 16, (24, 32)* bit PCM WAV.
   * 1-65535 channel audio.
   * * untested.
   * @param in Object to paste into the current stream.
   * @param where Sample number to paste it after.
  */
  void bliepdsp::paste (bliepdsp *in, const unsigned int where);

  /**
   * @brief Fade out the stream.
   *
   * Fade the end of the stream to zero.
   * 8, 16  bit PCM WAV.
   * 1-65535 channel audio.
   * @param out Object contains the result of the fade-out.
   * @param fadetime Number of milliseconds to fade out.
  */
  void bliepdsp::fade_out (bliepdsp *out, const unsigned int fadetime);

  /**
   * @brief Fade in the stream.
   *
   * Fades in the stream from zero.
   * 8, 16  bit PCM WAV.
   * 1-65535 channel audio.
   * @param out Object contains the result of the fade-in.
   * @param fadetime Number of milliseconds to fade in.
  */
  void bliepdsp::fade_in (bliepdsp *out, const unsigned int fadetime);
};
#endif
